@@ -1,7 +1,7 @@
 //! Domain specific language for the Lazy api.
 use crate::logical_plan::Context;
 use crate::prelude::*;
-use crate::utils::{has_expr, has_wildcard, output_name};
+use crate::utils::{has_expr, has_wildcard};
 use polars_core::prelude::*;
 
 #[cfg(feature = "temporal")]
@@ -16,6 +16,8 @@ use std::{
 // reexport the lazy method
 pub use crate::frame::IntoLazy;
 use polars_core::frame::select::Selection;
+#[cfg(feature = "diff")]
+use polars_core::series::ops::NullBehavior;
 use polars_core::utils::get_supertype;
 
 /// A wrapper trait for any closure `Fn(Vec<Series>) -> Result<Series>`
@@ -91,7 +93,7 @@ impl<T> Deref for NoEq<T> {
 pub trait BinaryUdfOutputField: Send + Sync {
     fn get_field(
         &self,
-        _input_schema: &Schema,
+        input_schema: &Schema,
         cntxt: Context,
         field_a: &Field,
         field_b: &Field,
@@ -113,11 +115,70 @@ where
     }
 }
 
+pub trait FunctionOutputField: Send + Sync {
+    fn get_field(&self, input_schema: &Schema, cntxt: Context, fields: &[Field]) -> Field;
+}
+
+pub type GetOutput = NoEq<Arc<dyn FunctionOutputField>>;
+
+impl Default for GetOutput {
+    fn default() -> Self {
+        NoEq::new(Arc::new(
+            |_input_schema: &Schema, _cntxt: Context, fields: &[Field]| fields[0].clone(),
+        ))
+    }
+}
+
+impl GetOutput {
+    pub fn same_type() -> Self {
+        Default::default()
+    }
+
+    pub fn from_type(dt: DataType) -> Self {
+        NoEq::new(Arc::new(move |_: &Schema, _: Context, flds: &[Field]| {
+            Field::new(flds[0].name(), dt.clone())
+        }))
+    }
+
+    pub fn map_field<F: 'static + Fn(&Field) -> Field + Send + Sync>(f: F) -> Self {
+        NoEq::new(Arc::new(move |_: &Schema, _: Context, flds: &[Field]| {
+            f(&flds[0])
+        }))
+    }
+}
+
+impl<F> FunctionOutputField for F
+where
+    F: Fn(&Schema, Context, &[Field]) -> Field + Send + Sync,
+{
+    fn get_field(&self, input_schema: &Schema, cntxt: Context, fields: &[Field]) -> Field {
+        self(input_schema, cntxt, fields)
+    }
+}
+
+#[derive(Copy, Clone, Debug, PartialEq)]
+pub enum ApplyOptions {
+    /// Collect groups to a list and apply the function over the groups.
+    /// This can be important in aggregation context.
+    ApplyGroups,
+    // collect groups to a list and then apply
+    ApplyList,
+    // do not collect before apply
+    ApplyFlat,
+}
+
+#[derive(Copy, Clone, Debug, PartialEq)]
+pub struct WindowOptions {
+    /// Explode the aggregated list and just do a hstack instead of a join
+    /// this requires the groups to be sorted to make any sense
+    pub(crate) explode: bool,
+}
+
 #[derive(Clone, Copy, PartialEq, Debug)]
 pub struct FunctionOptions {
-    /// Collect groups to a list before applying a function.
+    /// Collect groups to a list and apply the function over the groups.
     /// This can be important in aggregation context.
-    pub(crate) collect_groups: bool,
+    pub(crate) collect_groups: ApplyOptions,
     /// There can be two ways of expanding wildcards:
     ///
     /// Say the schema is 'a', 'b' and there is a function f
@@ -130,6 +191,8 @@ pub struct FunctionOptions {
     ///     f('a'), f('b')
     ///
     /// setting this to true, will lead to behavior 1.
+    ///
+    /// this also accounts for regex expansion
     pub(crate) input_wildcard_expansion: bool,
 }
 
@@ -184,6 +247,7 @@ impl From<AggExpr> for Expr {
 pub enum Expr {
     Alias(Box<Expr>, Arc<String>),
     Column(Arc<String>),
+    Columns(Vec<String>),
     Literal(LiteralValue),
     BinaryExpr {
         left: Box<Expr>,
@@ -196,6 +260,7 @@ pub enum Expr {
     Cast {
         expr: Box<Expr>,
         data_type: DataType,
+        strict: bool,
     },
     Sort {
         expr: Box<Expr>,
@@ -224,7 +289,7 @@ pub enum Expr {
         /// function to apply
         function: NoEq<Arc<dyn SeriesUdf>>,
         /// output dtype of the function
-        output_type: Option<DataType>,
+        output_type: GetOutput,
         options: FunctionOptions,
     },
     Shift {
@@ -245,6 +310,7 @@ pub enum Expr {
         function: Box<Expr>,
         partition_by: Vec<Expr>,
         order_by: Option<Box<Expr>>,
+        options: WindowOptions,
     },
     Wildcard,
     Slice {
@@ -264,6 +330,11 @@ pub enum Expr {
     Exclude(Box<Expr>, Vec<Arc<String>>),
     /// Set root name as Alias
     KeepName(Box<Expr>),
+    SufPreFix {
+        is_suffix: bool,
+        value: String,
+        expr: Box<Expr>,
+    },
 }
 
 impl Expr {
@@ -284,6 +355,7 @@ impl fmt::Debug for Expr {
                 function,
                 partition_by,
                 order_by,
+                ..
             } => write!(
                 f,
                 "{:?} OVER (PARTITION BY {:?} ORDER BY {:?}",
@@ -295,7 +367,17 @@ impl fmt::Debug for Expr {
             Reverse(expr) => write!(f, "REVERSE {:?}", expr),
             Alias(expr, name) => write!(f, "{:?} AS {}", expr, name),
             Column(name) => write!(f, "{}", name),
-            Literal(v) => write!(f, "{:?}", v),
+            Literal(v) => {
+                match v {
+                    LiteralValue::Utf8(v) => {
+                        // dot breaks with debug fmt due to \"
+                        write!(f, "Utf8({})", v)
+                    }
+                    _ => {
+                        write!(f, "{:?}", v)
+                    }
+                }
+            }
             BinaryExpr { left, op, right } => write!(f, "[({:?}) {:?} ({:?})]", left, op, right),
             Not(expr) => write!(f, "NOT {:?}", expr),
             IsNull(expr) => write!(f, "{:?} IS NULL", expr),
@@ -333,7 +415,9 @@ impl fmt::Debug for Expr {
                     Quantile { expr, .. } => write!(f, "AGG QUANTILE {:?}", expr),
                 }
             }
-            Cast { expr, data_type } => write!(f, "CAST {:?} TO {:?}", expr, data_type),
+            Cast {
+                expr, data_type, ..
+            } => write!(f, "CAST {:?} TO {:?}", expr, data_type),
             Ternary {
                 predicate,
                 truthy,
@@ -356,6 +440,8 @@ impl fmt::Debug for Expr {
             Wildcard => write!(f, "*"),
             Exclude(column, names) => write!(f, "{:?}, EXCEPT {:?}", column, names),
             KeepName(e) => write!(f, "KEEP NAME {:?}", e),
+            SufPreFix { expr, .. } => write!(f, "SUF-PREFIX {:?}", expr),
+            Columns(names) => write!(f, "COLUMNS({:?})", names),
         }
     }
 }
@@ -398,7 +484,7 @@ pub struct WhenThen {
 
 /// Intermediate state of chain when then exprs.
 ///
-/// ```text,ignore
+/// ```ignore
 /// when(..).then(..)
 /// when(..).then(..)
 /// when(..).then(..)
@@ -624,9 +710,41 @@ impl Expr {
         AggExpr::AggGroups(Box::new(self)).into()
     }
 
+    /// Alias for explode
+    pub fn flatten(self) -> Self {
+        self.explode()
+    }
+
     /// Explode the utf8/ list column
     pub fn explode(self) -> Self {
-        Expr::Explode(Box::new(self))
+        let has_filter = has_expr(&self, |e| matches!(e, Expr::Filter { .. }));
+
+        // if we explode right after a window function we don't self join, but just flatten
+        // the expression
+        if let Expr::Window {
+            function,
+            partition_by,
+            order_by,
+            mut options,
+        } = self
+        {
+            if has_filter {
+                panic!("A Filter of a window function is not allowed in combination with explode/flatten.\
+                The resulting column may not fit the DataFrame/ or the groups
+                ")
+            }
+
+            options.explode = true;
+
+            Expr::Explode(Box::new(Expr::Window {
+                function,
+                partition_by,
+                order_by,
+                options,
+            }))
+        } else {
+            Expr::Explode(Box::new(self))
+        }
     }
 
     /// Slice the Series.
@@ -654,7 +772,7 @@ impl Expr {
         if has_expr(&self, |e| matches!(e, Expr::Wildcard)) {
             panic!("wildcard not supperted in unique expr");
         }
-        self.apply(|s: Series| s.unique(), None)
+        self.apply(|s: Series| s.unique(), GetOutput::same_type())
     }
 
     /// Get the first index of unique values of this expression.
@@ -664,7 +782,7 @@ impl Expr {
         }
         self.apply(
             |s: Series| s.arg_unique().map(|ca| ca.into_series()),
-            Some(DataType::UInt32),
+            GetOutput::from_type(DataType::UInt32),
         )
     }
 
@@ -675,15 +793,26 @@ impl Expr {
         }
         self.apply(
             move |s: Series| Ok(s.argsort(reverse).into_series()),
-            Some(DataType::UInt32),
+            GetOutput::from_type(DataType::UInt32),
         )
     }
 
+    /// Cast expression to another data type.
+    /// Throws an error if conversion had overflows
+    pub fn strict_cast(self, data_type: DataType) -> Self {
+        Expr::Cast {
+            expr: Box::new(self),
+            data_type,
+            strict: true,
+        }
+    }
+
     /// Cast expression to another data type.
     pub fn cast(self, data_type: DataType) -> Self {
         Expr::Cast {
             expr: Box::new(self),
             data_type,
+            strict: false,
         }
     }
 
@@ -719,7 +848,7 @@ impl Expr {
     ///
     /// It is the responsibility of the caller that the schema is correct by giving
     /// the correct output_type. If None given the output type of the input expr is used.
-    pub fn map<F>(self, function: F, output_type: Option<DataType>) -> Self
+    pub fn map<F>(self, function: F, output_type: GetOutput) -> Self
     where
         F: Fn(Series) -> Result<Series> + 'static + Send + Sync,
     {
@@ -730,7 +859,31 @@ impl Expr {
             function: NoEq::new(Arc::new(f)),
             output_type,
             options: FunctionOptions {
-                collect_groups: false,
+                collect_groups: ApplyOptions::ApplyFlat,
+                input_wildcard_expansion: false,
+            },
+        }
+    }
+
+    /// Apply a function/closure once the logical plan get executed.
+    ///
+    /// This function is very similar to [apply](Expr::apply), but differs in how it handles aggregations.
+    ///
+    ///  * `map` should be used for operations that are independent of groups, e.g. `multiply * 2`, or `raise to the power`
+    ///  * `apply` should be used for operations that work on a group of data. e.g. `sum`, `count`, etc.
+    ///  * `map_list` should be used when the function expects a list aggregated series.
+    pub fn map_list<F>(self, function: F, output_type: GetOutput) -> Self
+    where
+        F: Fn(Series) -> Result<Series> + 'static + Send + Sync,
+    {
+        let f = move |s: &mut [Series]| function(std::mem::take(&mut s[0]));
+
+        Expr::Function {
+            input: vec![self],
+            function: NoEq::new(Arc::new(f)),
+            output_type,
+            options: FunctionOptions {
+                collect_groups: ApplyOptions::ApplyList,
                 input_wildcard_expansion: false,
             },
         }
@@ -745,7 +898,7 @@ impl Expr {
     ///
     /// * `map` should be used for operations that are independent of groups, e.g. `multiply * 2`, or `raise to the power`
     /// * `apply` should be used for operations that work on a group of data. e.g. `sum`, `count`, etc.
-    pub fn apply<F>(self, function: F, output_type: Option<DataType>) -> Self
+    pub fn apply<F>(self, function: F, output_type: GetOutput) -> Self
     where
         F: Fn(Series) -> Result<Series> + 'static + Send + Sync,
     {
@@ -756,7 +909,7 @@ impl Expr {
             function: NoEq::new(Arc::new(f)),
             output_type,
             options: FunctionOptions {
-                collect_groups: true,
+                collect_groups: ApplyOptions::ApplyGroups,
                 input_wildcard_expansion: false,
             },
         }
@@ -767,7 +920,7 @@ impl Expr {
     pub fn is_finite(self) -> Self {
         self.map(
             |s: Series| s.is_finite().map(|ca| ca.into_series()),
-            Some(DataType::Boolean),
+            GetOutput::from_type(DataType::Boolean),
         )
     }
 
@@ -776,7 +929,7 @@ impl Expr {
     pub fn is_infinite(self) -> Self {
         self.map(
             |s: Series| s.is_infinite().map(|ca| ca.into_series()),
-            Some(DataType::Boolean),
+            GetOutput::from_type(DataType::Boolean),
         )
     }
 
@@ -785,7 +938,7 @@ impl Expr {
     pub fn is_nan(self) -> Self {
         self.map(
             |s: Series| s.is_nan().map(|ca| ca.into_series()),
-            Some(DataType::Boolean),
+            GetOutput::from_type(DataType::Boolean),
         )
     }
 
@@ -794,7 +947,7 @@ impl Expr {
     pub fn is_not_nan(self) -> Self {
         self.map(
             |s: Series| s.is_not_nan().map(|ca| ca.into_series()),
-            Some(DataType::Boolean),
+            GetOutput::from_type(DataType::Boolean),
         )
     }
 
@@ -806,24 +959,22 @@ impl Expr {
         }
     }
 
-    /// Shift the valus in the array by some period and fill the resulting empty values.
+    /// Shift the values in the array by some period and fill the resulting empty values.
     pub fn shift_and_fill(self, periods: i64, fill_value: Expr) -> Self {
-        // let name = output_name(&self).unwrap();
         // Note:
-        // The order of the then | otherwise is im
+        // The order of the then | otherwise is important
         if periods > 0 {
-            when(self.clone().map(
+            when(self.clone().apply(
                 move |s: Series| {
                     let ca: BooleanChunked = (0..s.len() as i64).map(|i| i >= periods).collect();
                     Ok(ca.into_series())
                 },
-                Some(DataType::Boolean),
+                GetOutput::from_type(DataType::Boolean),
             ))
             .then(self.shift(periods))
             .otherwise(fill_value)
-            // .alias(&name)
         } else {
-            when(self.clone().map(
+            when(self.clone().apply(
                 move |s: Series| {
                     let length = s.len() as i64;
                     // periods is negative, so subtraction.
@@ -831,42 +982,53 @@ impl Expr {
                     let ca: BooleanChunked = (0..length).map(|i| i < tipping_point).collect();
                     Ok(ca.into_series())
                 },
-                Some(DataType::Boolean),
+                GetOutput::from_type(DataType::Boolean),
             ))
             .then(self.shift(periods))
             .otherwise(fill_value)
-            // .alias(&name)
         }
     }
 
     /// Get an array with the cumulative sum computed at every element
-    pub fn cum_sum(self, reverse: bool) -> Self {
-        self.apply(move |s: Series| Ok(s.cum_sum(reverse)), None)
+    #[cfg_attr(docsrs, doc(cfg(feature = "cum_agg")))]
+    pub fn cumsum(self, reverse: bool) -> Self {
+        self.apply(
+            move |s: Series| Ok(s.cumsum(reverse)),
+            GetOutput::same_type(),
+        )
     }
 
     /// Get an array with the cumulative min computed at every element
-    pub fn cum_min(self, reverse: bool) -> Self {
-        self.apply(move |s: Series| Ok(s.cum_min(reverse)), None)
+    #[cfg_attr(docsrs, doc(cfg(feature = "cum_agg")))]
+    pub fn cummin(self, reverse: bool) -> Self {
+        self.apply(
+            move |s: Series| Ok(s.cummin(reverse)),
+            GetOutput::same_type(),
+        )
     }
 
     /// Get an array with the cumulative max computed at every element
-    pub fn cum_max(self, reverse: bool) -> Self {
-        self.apply(move |s: Series| Ok(s.cum_max(reverse)), None)
+    #[cfg_attr(docsrs, doc(cfg(feature = "cum_agg")))]
+    pub fn cummax(self, reverse: bool) -> Self {
+        self.apply(
+            move |s: Series| Ok(s.cummax(reverse)),
+            GetOutput::same_type(),
+        )
     }
 
     /// Fill missing value with next non-null.
     pub fn backward_fill(self) -> Self {
         self.apply(
-            move |s: Series| s.fill_none(FillNoneStrategy::Backward),
-            None,
+            move |s: Series| s.fill_null(FillNullStrategy::Backward),
+            GetOutput::same_type(),
         )
     }
 
     /// Fill missing value with previous non-null.
     pub fn forward_fill(self) -> Self {
         self.apply(
-            move |s: Series| s.fill_none(FillNoneStrategy::Forward),
-            None,
+            move |s: Series| s.fill_null(FillNullStrategy::Forward),
+            GetOutput::same_type(),
         )
     }
 
@@ -874,7 +1036,7 @@ impl Expr {
     #[cfg(feature = "round_series")]
     #[cfg_attr(docsrs, doc(cfg(feature = "round_series")))]
     pub fn round(self, decimals: u32) -> Self {
-        self.apply(move |s: Series| s.round(decimals), None)
+        self.map(move |s: Series| s.round(decimals), GetOutput::same_type())
     }
 
     /// Apply window function over a subgroup.
@@ -941,16 +1103,31 @@ impl Expr {
             function: Box::new(self),
             partition_by,
             order_by: None,
+            options: WindowOptions { explode: false },
         }
     }
 
-    /// Shift the values in the array by some period. See [the eager implementation](polars_core::series::SeriesTrait::fill_none).
-    pub fn fill_none(self, fill_value: Expr) -> Self {
-        let name = output_name(&self).unwrap();
-        when(self.is_null())
-            .then(fill_value)
-            .otherwise(col(&*name))
-            .alias(&*name)
+    /// Shift the values in the array by some period. See [the eager implementation](polars_core::series::SeriesTrait::fill_null).
+    pub fn fill_null(self, fill_value: Expr) -> Self {
+        map_binary_lazy_field(
+            self,
+            fill_value,
+            |a, b| {
+                if a.null_count() == 0 {
+                    Ok(a)
+                } else {
+                    let st = get_supertype(a.dtype(), b.dtype())?;
+                    let a = a.cast_with_dtype(&st)?;
+                    let b = b.cast_with_dtype(&st)?;
+                    let mask = a.is_not_null();
+                    a.zip_with_same_type(&mask, &b)
+                }
+            },
+            |_schema, _ctx, a, b| {
+                let st = get_supertype(a.data_type(), b.data_type()).unwrap();
+                Some(Field::new(a.name(), st))
+            },
+        )
     }
     /// Count the values of the Series
     /// or
@@ -993,7 +1170,10 @@ impl Expr {
 
     /// Raise expression to the power `exponent`
     pub fn pow(self, exponent: f64) -> Self {
-        self.map(move |s: Series| s.pow(exponent), Some(DataType::Float64))
+        self.map(
+            move |s: Series| s.pow(exponent),
+            GetOutput::from_type(DataType::Float64),
+        )
     }
 
     /// Filter a single column
@@ -1032,51 +1212,72 @@ impl Expr {
     #[cfg(feature = "temporal")]
     pub fn year(self) -> Expr {
         let function = move |s: Series| s.year().map(|ca| ca.into_series());
-        self.map(function, Some(DataType::UInt32))
+        self.map(function, GetOutput::from_type(DataType::UInt32))
     }
 
     /// Get the month of a Date32/Date64
     #[cfg(feature = "temporal")]
     pub fn month(self) -> Expr {
         let function = move |s: Series| s.month().map(|ca| ca.into_series());
-        self.map(function, Some(DataType::UInt32))
+        self.map(function, GetOutput::from_type(DataType::UInt32))
     }
+    /// Extract the week from the underlying Date representation.
+    /// Can be performed on Date32 and Date64
+
+    /// Returns the ISO week number starting from 1.
+    /// The return value ranges from 1 to 53. (The last week of year differs by years.)
+    #[cfg(feature = "temporal")]
+    pub fn week(self) -> Expr {
+        let function = move |s: Series| s.week().map(|ca| ca.into_series());
+        self.map(function, GetOutput::from_type(DataType::UInt32))
+    }
+
+    /// Extract the week day from the underlying Date representation.
+    /// Can be performed on Date32 and Date64.
+
+    /// Returns the weekday number where monday = 0 and sunday = 6
+    #[cfg(feature = "temporal")]
+    pub fn weekday(self) -> Expr {
+        let function = move |s: Series| s.weekday().map(|ca| ca.into_series());
+        self.map(function, GetOutput::from_type(DataType::UInt32))
+    }
+
     /// Get the month of a Date32/Date64
     #[cfg(feature = "temporal")]
     pub fn day(self) -> Expr {
         let function = move |s: Series| s.day().map(|ca| ca.into_series());
-        self.map(function, Some(DataType::UInt32))
+        self.map(function, GetOutput::from_type(DataType::UInt32))
     }
     /// Get the ordinal_day of a Date32/Date64
     #[cfg(feature = "temporal")]
     pub fn ordinal_day(self) -> Expr {
         let function = move |s: Series| s.ordinal_day().map(|ca| ca.into_series());
-        self.map(function, Some(DataType::UInt32))
+        self.map(function, GetOutput::from_type(DataType::UInt32))
     }
     /// Get the hour of a Date64/Time64
     #[cfg(feature = "temporal")]
     pub fn hour(self) -> Expr {
         let function = move |s: Series| s.hour().map(|ca| ca.into_series());
-        self.map(function, Some(DataType::UInt32))
+        self.map(function, GetOutput::from_type(DataType::UInt32))
     }
     /// Get the minute of a Date64/Time64
     #[cfg(feature = "temporal")]
     pub fn minute(self) -> Expr {
         let function = move |s: Series| s.minute().map(|ca| ca.into_series());
-        self.map(function, Some(DataType::UInt32))
+        self.map(function, GetOutput::from_type(DataType::UInt32))
     }
 
     /// Get the second of a Date64/Time64
     #[cfg(feature = "temporal")]
     pub fn second(self) -> Expr {
         let function = move |s: Series| s.second().map(|ca| ca.into_series());
-        self.map(function, Some(DataType::UInt32))
+        self.map(function, GetOutput::from_type(DataType::UInt32))
     }
     /// Get the nanosecond of a Time64
     #[cfg(feature = "temporal")]
     pub fn nanosecond(self) -> Expr {
         let function = move |s: Series| s.nanosecond().map(|ca| ca.into_series());
-        self.map(function, Some(DataType::UInt32))
+        self.map(function, GetOutput::from_type(DataType::UInt32))
     }
 
     /// Sort this column by the ordering of another column.
@@ -1109,9 +1310,9 @@ impl Expr {
     #[allow(clippy::wrong_self_convention)]
     /// Get a mask of the first unique value.
     pub fn is_first(self) -> Expr {
-        self.map(
+        self.apply(
             |s| s.is_first().map(|ca| ca.into_series()),
-            Some(DataType::Boolean),
+            GetOutput::from_type(DataType::Boolean),
         )
     }
 
@@ -1127,9 +1328,12 @@ impl Expr {
 
     #[cfg(feature = "mode")]
     #[cfg_attr(docsrs, doc(cfg(feature = "mode")))]
-    /// Compute the mode(s) of this column. These is the most occurring value.
+    /// Compute the mode(s) of this column. This is the most occurring value.
     pub fn mode(self) -> Expr {
-        self.map(|s| s.mode().map(|ca| ca.into_series()), None)
+        self.apply(
+            |s| s.mode().map(|ca| ca.into_series()),
+            GetOutput::same_type(),
+        )
     }
 
     /// Keep the original root name
@@ -1150,6 +1354,24 @@ impl Expr {
         Expr::KeepName(Box::new(self))
     }
 
+    /// Add a suffix to the root column name.
+    pub fn suffix(self, suffix: &str) -> Expr {
+        Expr::SufPreFix {
+            is_suffix: true,
+            value: suffix.to_string(),
+            expr: Box::new(self),
+        }
+    }
+
+    /// Add a prefix to the root column name.
+    pub fn prefix(self, prefix: &str) -> Expr {
+        Expr::SufPreFix {
+            is_suffix: false,
+            value: prefix.to_string(),
+            expr: Box::new(self),
+        }
+    }
+
     /// Exclude a column from a wildcard selection
     ///
     /// # Example
@@ -1177,6 +1399,142 @@ impl Expr {
             .collect();
         Expr::Exclude(Box::new(self), v)
     }
+
+    // Interpolate None values
+    #[cfg(feature = "interpolate")]
+    #[cfg_attr(docsrs, doc(cfg(feature = "interpolate")))]
+    pub fn interpolate(self) -> Expr {
+        self.apply(|s| Ok(s.interpolate()), GetOutput::same_type())
+    }
+
+    /// Apply a rolling min See:
+    /// [ChunkedArray::rolling_min](polars::prelude::ChunkWindow::rolling_min).
+    #[cfg_attr(docsrs, doc(cfg(feature = "rolling_window")))]
+    #[cfg(feature = "rolling_window")]
+    pub fn rolling_min(
+        self,
+        window_size: u32,
+        weight: Option<&[f64]>,
+        ignore_null: bool,
+        min_periods: u32,
+    ) -> Expr {
+        let weight = weight.map(|v| v.to_vec());
+        self.apply(
+            move |s| s.rolling_min(window_size, weight.as_deref(), ignore_null, min_periods),
+            GetOutput::same_type(),
+        )
+    }
+
+    /// Apply a rolling max See:
+    /// [ChunkedArray::rolling_max](polars::prelude::ChunkWindow::rolling_max).
+    #[cfg_attr(docsrs, doc(cfg(feature = "rolling_window")))]
+    #[cfg(feature = "rolling_window")]
+    pub fn rolling_max(
+        self,
+        window_size: u32,
+        weight: Option<&[f64]>,
+        ignore_null: bool,
+        min_periods: u32,
+    ) -> Expr {
+        let weight = weight.map(|v| v.to_vec());
+        self.apply(
+            move |s| s.rolling_max(window_size, weight.as_deref(), ignore_null, min_periods),
+            GetOutput::same_type(),
+        )
+    }
+
+    /// Apply a rolling mean See:
+    /// [ChunkedArray::rolling_mean](polars::prelude::ChunkWindow::rolling_mean).
+    #[cfg_attr(docsrs, doc(cfg(feature = "rolling_window")))]
+    #[cfg(feature = "rolling_window")]
+    pub fn rolling_mean(
+        self,
+        window_size: u32,
+        weight: Option<&[f64]>,
+        ignore_null: bool,
+        min_periods: u32,
+    ) -> Expr {
+        let weight = weight.map(|v| v.to_vec());
+        self.apply(
+            move |s| s.rolling_mean(window_size, weight.as_deref(), ignore_null, min_periods),
+            GetOutput::same_type(),
+        )
+    }
+
+    /// Apply a rolling sum See:
+    /// [ChunkedArray::rolling_sum](polars::prelude::ChunkWindow::rolling_sum).
+    #[cfg_attr(docsrs, doc(cfg(feature = "rolling_window")))]
+    #[cfg(feature = "rolling_window")]
+    pub fn rolling_sum(
+        self,
+        window_size: u32,
+        weight: Option<&[f64]>,
+        ignore_null: bool,
+        min_periods: u32,
+    ) -> Expr {
+        let weight = weight.map(|v| v.to_vec());
+        self.apply(
+            move |s| s.rolling_sum(window_size, weight.as_deref(), ignore_null, min_periods),
+            GetOutput::same_type(),
+        )
+    }
+
+    #[cfg_attr(docsrs, doc(cfg(feature = "rolling_window")))]
+    #[cfg(feature = "rolling_window")]
+    /// Apply a custom function over a rolling/ moving window of the array.
+    /// This has quite some dynamic dispatch, so prefer rolling_min, max, mean, sum over this.
+    pub fn rolling_apply(
+        self,
+        window_size: usize,
+        f: Arc<dyn Fn(&Series) -> Series + Send + Sync>,
+    ) -> Expr {
+        self.apply(
+            move |s| s.rolling_apply(window_size, f.as_ref()),
+            GetOutput::same_type(),
+        )
+    }
+
+    #[cfg(feature = "rank")]
+    #[cfg_attr(docsrs, doc(cfg(feature = "rank")))]
+    pub fn rank(self, method: RankMethod) -> Expr {
+        self.apply(
+            move |s| Ok(s.rank(method)),
+            GetOutput::map_field(move |fld| match method {
+                RankMethod::Average => Field::new(fld.name(), DataType::Float32),
+                _ => Field::new(fld.name(), DataType::UInt32),
+            }),
+        )
+    }
+
+    #[cfg(feature = "diff")]
+    #[cfg_attr(docsrs, doc(cfg(feature = "diff")))]
+    pub fn diff(self, n: usize, null_behavior: NullBehavior) -> Expr {
+        self.apply(
+            move |s| Ok(s.diff(n, null_behavior)),
+            GetOutput::same_type(),
+        )
+    }
+
+    #[cfg(feature = "moment")]
+    #[cfg_attr(docsrs, doc(cfg(feature = "moment")))]
+    pub fn skew(self, bias: bool) -> Expr {
+        self.apply(
+            move |s| s.skew(bias).map(|opt_v| Series::new(s.name(), &[opt_v])),
+            GetOutput::from_type(DataType::Float64),
+        )
+    }
+
+    #[cfg(feature = "moment")]
+    #[cfg_attr(docsrs, doc(cfg(feature = "moment")))]
+    pub fn kurtosis(self, fisher: bool, bias: bool) -> Expr {
+        self.apply(
+            move |s| {
+                s.kurtosis(fisher, bias)
+                    .map(|opt_v| Series::new(s.name(), &[opt_v]))
+            },
+            GetOutput::from_type(DataType::Float64),
+        )
+    }
 }
 
 /// Create a Column Expression based on a column name.
@@ -1209,6 +1567,11 @@ pub fn col(name: &str) -> Expr {
     }
 }
 
+/// Select multiple columns by name
+pub fn cols(names: Vec<String>) -> Expr {
+    Expr::Columns(names)
+}
+
 /// Count the number of values in this Expression.
 pub fn count(name: &str) -> Expr {
     match name {
@@ -1304,12 +1667,13 @@ where
             Ok(acc)
         }) as Arc<dyn SeriesUdf>);
 
+        // Todo! make sure that output type is correct
         Expr::Function {
             input: exprs,
             function,
-            output_type: None,
+            output_type: GetOutput::same_type(),
             options: FunctionOptions {
-                collect_groups: false,
+                collect_groups: ApplyOptions::ApplyFlat,
                 input_wildcard_expansion: true,
             },
         }
@@ -1465,6 +1829,7 @@ pub fn cast(expr: Expr, data_type: DataType) -> Expr {
     Expr::Cast {
         expr: Box::new(expr),
         data_type,
+        strict: false,
     }
 }
 
