@@ -28,9 +28,14 @@ pub(crate) fn to_aexpr(expr: Expr, arena: &mut Arena<AExpr>) -> Node {
         Expr::IsNotNull(e) => AExpr::IsNotNull(to_aexpr(*e, arena)),
         Expr::IsNull(e) => AExpr::IsNull(to_aexpr(*e, arena)),
 
-        Expr::Cast { expr, data_type } => AExpr::Cast {
+        Expr::Cast {
+            expr,
+            data_type,
+            strict,
+        } => AExpr::Cast {
             expr: to_aexpr(*expr, arena),
             data_type,
+            strict,
         },
         Expr::Take { expr, idx } => AExpr::Take {
             expr: to_aexpr(*expr, arena),
@@ -89,12 +94,12 @@ pub(crate) fn to_aexpr(expr: Expr, arena: &mut Arena<AExpr>) -> Node {
             input,
             function,
             output_type,
-            collect_groups,
+            options,
         } => AExpr::Function {
             input: to_aexprs(input, arena),
             function,
             output_type,
-            collect_groups,
+            options,
         },
         Expr::BinaryFunction {
             input_a,
@@ -115,10 +120,12 @@ pub(crate) fn to_aexpr(expr: Expr, arena: &mut Arena<AExpr>) -> Node {
             function,
             partition_by,
             order_by,
+            options,
         } => AExpr::Window {
             function: to_aexpr(*function, arena),
             partition_by: to_aexprs(partition_by, arena),
             order_by: order_by.map(|ob| to_aexpr(*ob, arena)),
+            options,
         },
         Expr::Slice {
             input,
@@ -130,7 +137,10 @@ pub(crate) fn to_aexpr(expr: Expr, arena: &mut Arena<AExpr>) -> Node {
             length,
         },
         Expr::Wildcard => AExpr::Wildcard,
-        Expr::Except(input) => AExpr::Except(to_aexpr(*input, arena)),
+        Expr::KeepName(_) => panic!("no keep_name expected at this point"),
+        Expr::Exclude(_, _) => panic!("no exclude expected at this point"),
+        Expr::SufPreFix { .. } => panic!("no `suffix/prefix` expected at this point"),
+        Expr::Columns { .. } => panic!("no `columns` expected at this point"),
     };
     arena.add(v)
 }
@@ -171,32 +181,18 @@ pub(crate) fn to_alp(
         LogicalPlan::CsvScan {
             path,
             schema,
-            has_header,
-            delimiter,
-            ignore_errors,
-            skip_rows,
-            stop_after_n_rows,
-            with_columns,
+            options,
             predicate,
             aggregate,
-            cache,
-            low_memory,
         } => ALogicalPlan::CsvScan {
             path,
             schema,
-            has_header,
-            delimiter,
-            ignore_errors,
-            skip_rows,
-            stop_after_n_rows,
-            with_columns,
+            options,
             predicate: predicate.map(|expr| to_aexpr(expr, expr_arena)),
             aggregate: aggregate
                 .into_iter()
                 .map(|expr| to_aexpr(expr, expr_arena))
                 .collect(),
-            cache,
-            low_memory,
         },
         #[cfg(feature = "parquet")]
         LogicalPlan::ParquetScan {
@@ -287,6 +283,7 @@ pub(crate) fn to_alp(
             aggs,
             schema,
             apply,
+            maintain_order,
         } => {
             let i = to_alp(*input, expr_arena, lp_arena);
             let aggs_new = aggs.into_iter().map(|x| to_aexpr(x, expr_arena)).collect();
@@ -301,6 +298,7 @@ pub(crate) fn to_alp(
                 aggs: aggs_new,
                 schema,
                 apply,
+                maintain_order,
             }
         }
         LogicalPlan::Join {
@@ -416,11 +414,16 @@ pub(crate) fn node_to_exp(node: Node, expr_arena: &Arena<AExpr>) -> Expr {
             let exp = node_to_exp(expr, expr_arena);
             Expr::IsNull(Box::new(exp))
         }
-        AExpr::Cast { expr, data_type } => {
+        AExpr::Cast {
+            expr,
+            data_type,
+            strict,
+        } => {
             let exp = node_to_exp(expr, expr_arena);
             Expr::Cast {
                 expr: Box::new(exp),
                 data_type,
+                strict,
             }
         }
         AExpr::Sort { expr, reverse } => {
@@ -544,12 +547,12 @@ pub(crate) fn node_to_exp(node: Node, expr_arena: &Arena<AExpr>) -> Expr {
             input,
             function,
             output_type,
-            collect_groups,
+            options,
         } => Expr::Function {
             input: nodes_to_exprs(&input, expr_arena),
             function,
             output_type,
-            collect_groups,
+            options,
         },
         AExpr::BinaryFunction {
             input_a,
@@ -566,6 +569,7 @@ pub(crate) fn node_to_exp(node: Node, expr_arena: &Arena<AExpr>) -> Expr {
             function,
             partition_by,
             order_by,
+            options,
         } => {
             let function = Box::new(node_to_exp(function, expr_arena));
             let partition_by = nodes_to_exprs(&partition_by, expr_arena);
@@ -574,6 +578,7 @@ pub(crate) fn node_to_exp(node: Node, expr_arena: &Arena<AExpr>) -> Expr {
                 function,
                 partition_by,
                 order_by,
+                options,
             }
         }
         AExpr::Slice {
@@ -586,7 +591,6 @@ pub(crate) fn node_to_exp(node: Node, expr_arena: &Arena<AExpr>) -> Expr {
             length,
         },
         AExpr::Wildcard => Expr::Wildcard,
-        AExpr::Except(node) => Expr::Except(Box::new(node_to_exp(node, expr_arena))),
     }
 }
 
@@ -623,29 +627,15 @@ pub(crate) fn node_to_lp(
         ALogicalPlan::CsvScan {
             path,
             schema,
-            has_header,
-            delimiter,
-            ignore_errors,
-            skip_rows,
-            stop_after_n_rows,
-            with_columns,
+            options,
             predicate,
             aggregate,
-            cache,
-            low_memory,
         } => LogicalPlan::CsvScan {
             path,
             schema,
-            has_header,
-            delimiter,
-            ignore_errors,
-            skip_rows,
-            stop_after_n_rows,
-            with_columns,
+            options,
             predicate: predicate.map(|n| node_to_exp(n, expr_arena)),
             aggregate: nodes_to_exprs(&aggregate, expr_arena),
-            cache,
-            low_memory,
         },
         #[cfg(feature = "parquet")]
         ALogicalPlan::ParquetScan {
@@ -730,6 +720,7 @@ pub(crate) fn node_to_lp(
             aggs,
             schema,
             apply,
+            maintain_order,
         } => {
             let i = node_to_lp(input, expr_arena, lp_arena);
 
@@ -739,6 +730,7 @@ pub(crate) fn node_to_lp(
                 aggs: nodes_to_exprs(&aggs, expr_arena),
                 schema,
                 apply,
+                maintain_order,
             }
         }
         ALogicalPlan::Join {
