@@ -52,10 +52,43 @@ impl PhysicalExpr for BinaryExpr {
     }
 
     fn evaluate(&self, df: &DataFrame, state: &ExecutionState) -> Result<Series> {
-        let lhs = self.left.evaluate(df, state)?;
-        let rhs = self.right.evaluate(df, state)?;
-        apply_operator(&lhs, &rhs, self.op)
+        let (lhs, rhs) = POOL.install(|| {
+            rayon::join(
+                || self.left.evaluate(df, state),
+                || self.right.evaluate(df, state),
+            )
+        });
+        apply_operator(&lhs?, &rhs?, self.op)
     }
+
+    #[allow(clippy::ptr_arg)]
+    fn evaluate_on_groups<'a>(
+        &self,
+        df: &DataFrame,
+        groups: &'a GroupTuples,
+        state: &ExecutionState,
+    ) -> Result<AggregationContext<'a>> {
+        let (result_a, result_b) = POOL.install(|| {
+            rayon::join(
+                || self.left.evaluate_on_groups(df, groups, state),
+                || self.right.evaluate_on_groups(df, groups, state),
+            )
+        });
+        let mut ac_l = result_a?;
+        let ac_r = result_b?;
+
+        if !ac_l.can_combine(&ac_r) {
+            return Err(PolarsError::InvalidOperation(
+                "\
+            cannot combine this binary expression, the groups do not match"
+                    .into(),
+            ));
+        }
+        let out = apply_operator(ac_l.flat().as_ref(), ac_r.flat().as_ref(), self.op)?;
+        ac_l.combine_groups(ac_r).with_series(out);
+        Ok(ac_l)
+    }
+
     fn to_field(&self, _input_schema: &Schema) -> Result<Field> {
         todo!()
     }
@@ -73,29 +106,6 @@ impl PhysicalAggregation for BinaryExpr {
         state: &ExecutionState,
     ) -> Result<Option<Series>> {
         match (self.left.as_agg_expr(), self.right.as_agg_expr()) {
-            (Ok(left), Err(_)) => {
-                let (opt_agg, rhs) = POOL.install(|| {
-                    rayon::join(
-                        || left.aggregate(df, groups, state),
-                        || self.right.evaluate(df, state),
-                    )
-                });
-                opt_agg?
-                    .map(|agg| apply_operator(&agg, &rhs?, self.op))
-                    .transpose()
-            }
-            (Err(_), Ok(right)) => {
-                let (opt_agg, lhs) = POOL.install(|| {
-                    rayon::join(
-                        || right.aggregate(df, groups, state),
-                        || self.left.evaluate(df, state),
-                    )
-                });
-
-                opt_agg?
-                    .map(|agg| apply_operator(&lhs?, &agg, self.op))
-                    .transpose()
-            }
             (Ok(left), Ok(right)) => {
                 let (left_agg, right_agg) = POOL.install(|| {
                     rayon::join(
@@ -108,8 +118,14 @@ impl PhysicalAggregation for BinaryExpr {
                     .and_then(|left| right_agg.map(|right| apply_operator(&left, &right, self.op)))
                     .transpose()
             }
-            (_, _) => Err(PolarsError::Other(
-                "both expressions could not be used in an aggregation context.".into(),
+            (_, _) => Err(PolarsError::ComputeError(
+                format!(
+                    "this binary expression is not an aggregation: {:?}
+                pherhaps you should add an aggregation like, '.sum()', '.min()', '.mean()', etc.
+                if you really want to collect this binary expression, use `.list()`",
+                    self.expr
+                )
+                .into(),
             )),
         }
     }
