@@ -4,6 +4,7 @@ use crate::prelude::*;
 use polars_core::frame::groupby::GroupTuples;
 use polars_core::prelude::*;
 use polars_core::utils::NoNull;
+use std::borrow::Cow;
 use std::ops::Deref;
 
 pub struct LiteralExpr(pub LiteralValue, Expr);
@@ -57,7 +58,7 @@ impl PhysicalExpr for LiteralExpr {
                 }
                 DataType::UInt32 => {
                     if *low >= 0 || *high <= u32::MAX as i64 {
-                        return Err(PolarsError::Other(
+                        return Err(PolarsError::ComputeError(
                             "range not within bounds of u32 type".into(),
                         ));
                     }
@@ -84,6 +85,17 @@ impl PhysicalExpr for LiteralExpr {
         Ok(s)
     }
 
+    #[allow(clippy::ptr_arg)]
+    fn evaluate_on_groups<'a>(
+        &self,
+        df: &DataFrame,
+        groups: &'a GroupTuples,
+        state: &ExecutionState,
+    ) -> Result<AggregationContext<'a>> {
+        let s = self.evaluate(df, state)?;
+        Ok(AggregationContext::new(s, Cow::Borrowed(groups)))
+    }
+
     fn to_field(&self, _input_schema: &Schema) -> Result<Field> {
         use LiteralValue::*;
         let name = "literal";
