@@ -1,5 +1,6 @@
 use crate::physical_plan::state::ExecutionState;
 use crate::prelude::*;
+use polars_core::frame::groupby::GroupTuples;
 use polars_core::prelude::*;
 use std::sync::Arc;
 
@@ -9,14 +10,8 @@ impl NotExpr {
     pub fn new(physical_expr: Arc<dyn PhysicalExpr>, expr: Expr) -> Self {
         Self(physical_expr, expr)
     }
-}
-impl PhysicalExpr for NotExpr {
-    fn as_expression(&self) -> &Expr {
-        &self.1
-    }
 
-    fn evaluate(&self, df: &DataFrame, state: &ExecutionState) -> Result<Series> {
-        let series = self.0.evaluate(df, state)?;
+    fn finish(&self, series: Series) -> Result<Series> {
         if let Ok(ca) = series.bool() {
             Ok((!ca).into_series())
         } else {
@@ -25,6 +20,30 @@ impl PhysicalExpr for NotExpr {
             ))
         }
     }
+}
+impl PhysicalExpr for NotExpr {
+    fn as_expression(&self) -> &Expr {
+        &self.1
+    }
+
+    fn evaluate(&self, df: &DataFrame, state: &ExecutionState) -> Result<Series> {
+        let series = self.0.evaluate(df, state)?;
+        self.finish(series)
+    }
+    #[allow(clippy::ptr_arg)]
+    fn evaluate_on_groups<'a>(
+        &self,
+        df: &DataFrame,
+        groups: &'a GroupTuples,
+        state: &ExecutionState,
+    ) -> Result<AggregationContext<'a>> {
+        let mut ac = self.0.evaluate_on_groups(df, groups, state)?;
+        let s = ac.flat().into_owned();
+        ac.with_series(self.finish(s)?);
+
+        Ok(ac)
+    }
+
     fn to_field(&self, _input_schema: &Schema) -> Result<Field> {
         Ok(Field::new("not", DataType::Boolean))
     }
